/*
  dlib is a machine learning toolkit that can be used with C++ or Python
  with support also for "deep" neural networks (dnn).

  Example programm (MNIST): http://dlib.net/dnn_introduction_ex.cpp.html

  In this example we take our simple 1-layer fully connected network as seen
  natively in Octave and run it using dlib ( http://dlib.net ).

  While dlib still uses gradient descent to compute variation in weights during
  the training (actually stochastic gradient descent), it has a more sofisticated
  convergence criterium to understand when it should stop training.

  It uses part of the training data as validation set (not to be confused with
  our test set) and measured the errors ("loss") during training. In the
  words of the author:

    // By default, it runs SGD with our specified learning rate until the loss stops
    // decreasing.  Then it reduces the learning rate by a factor of 10 and continues
    // running until the loss stops decreasing again.  It will keep doing this until
    // the learning rate has dropped below the min learning rate defined above or the
    // maximum number of epochs as been executed (defaulted to 10000).

  One small difference is we use relu as the activation function instead of sigmoid
  (for some reason dlib's training algo like relu more it seems, however sigmoid is
  also available, just use sig<>).

  Artificially limiting to 5 epochs, finishes in 8 seconds on my notebook
  and gives an accuracy of 84.43 % in line what we got with 02-basic-fc-nn-noframework-octave.

  Without limiting the epochs, dlib ran up to epoch 182 and took 300 seconds on my notebook
  and correct detection rate improved to 89.49 %.

  Note dlib also can use all 8 available cores.

*/


#include <iostream>
#include <dlib/matrix.h>
#include <dlib/dnn.h>
#include "load_images.h"

using std::cout;
using namespace dlib;

int main() {

    // --- load test data as prepared by 01-preprocess-images --------------------------------

    std::vector<matrix<uint8_t>>      train_image;
    std::vector<unsigned long>        train_label;
    std::vector<matrix<uint8_t>>      test_image;
    std::vector<unsigned long>        test_label;

    const int TRAIN_COUNT = 39209;
    load_images("../train_image.bin", "../train_label.bin", TRAIN_COUNT, train_image, train_label);

    const int TEST_COUNT = 12630;
    load_images("../test_image.bin", "../test_label.bin", TEST_COUNT, test_image, test_label);

    // --- define the net and train it -------------------------------------------------------

    using net_type = loss_multiclass_log<fc<43,
                     relu<fc<1000,
                     relu<input<matrix<uint8_t>>>>>>>;

    net_type net;

    dnn_trainer<net_type> trainer(net);
    trainer.set_learning_rate(0.01);
    trainer.set_min_learning_rate(0.00001);
    trainer.set_mini_batch_size(128);
    trainer.be_verbose();

    // artificially limit training to 5 epochs

    trainer.set_max_num_epochs(5);

    if (trainer.get_max_num_epochs() < 10000) {
        cout << "WARNING, max epochs hat been lowered artificially, this is not dlib's default\n";
    }

    trainer.train(train_image, train_label);
    net.clean();

    // --- score the net using the test images -----------------------------------------------

    std::vector<unsigned long> estimates = net(test_image);
    int num_right = 0;
    int num_wrong = 0;
    for (int i = 0; i < test_image.size(); ++i) {
        if (estimates[i] == test_label[i]) {
            num_right++;
        } else {
            num_wrong++;
        }
    }
    cout << "test correct:  " << num_right << "\n";
    cout << "test wrong:    " << num_wrong << "\n";
    cout << "test accuracy: " << num_right / (double)(num_right + num_wrong) << "\n";

    return 0;
}
